# ========================================================================
# Copyright 2018 Emory University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========================================================================
from typing import Any

import pytest

from elit.component import Component

__author__ = "Gary Lai"


def test_component():
    class TestComponent(Component):
        pass

    with pytest.raises(TypeError):
        TestComponent()


def test_space_tokenizer(space_tokenizer):
    result = space_tokenizer.decode("Hello, world")
    assert result['tok'] == ['Hello,', 'world']
    assert result['off'] == [(0, 6), (7, 12)]
