
# THIS FILE IS GENERATED FROM SCIPY SETUP.PY
short_version = '0.1.25'
version = '0.1.25'
full_version = '0.1.25.dev1535232511'
dev_version = '1535232511'
release = False
if not release:
    version = full_version
